""" reduction class for spectrograph """

import os
import sys
import shutil
import copy
import numpy as np
import matplotlib.pyplot as plt

from glob import glob, iglob
from tabulate import tabulate
from matplotlib.offsetbox import AnchoredText

from .tools import load_fio, load_tiff, flatten, peak_fit, binning


class spectrograph:

    def __init__(
        self,
        exp,
        detfac=0,
        threshold=15,
        cutoff=2000,
        roix=None,
        roih=None,
        roic=None,
        bias_correct=True,
        vmax_average=100,
        vmax_single=100,
        vmin=1,
        detector_type="greateyes",
        datdir_remote="/gpfs/current/raw",
        datdir_local="raw",
        savedir="processed",
    ):

        # if ROI is not given, use detector limits
        if roix is None:
            roix = [0, 2048]
        if roih is None:
            if detector_type == "greateyes":
                roih = 2064
            if detector_type == "andor":
                roih = 2048
        if roic is None:
            roic = roih // 2

        self.exp = exp
        self.detfac = detfac
        self.threshold = threshold
        self.cutoff = cutoff
        self.roix = roix
        self.roih = roih
        self.roic = roic
        self.bias_correct = bias_correct
        self.vmax_average = vmax_average
        self.vmax_single = vmax_single
        self.vmin = vmin
        self.detector_type = detector_type
        self.datdir = datdir_remote
        self.localdir = datdir_local
        self.savedir = savedir

        os.makedirs(self.localdir, exist_ok=True)
        os.makedirs(self.savedir, exist_ok=True)

        self.runs = {}

    def extract(self, run_nos):
        """ Extract run data and information
        Imports detector tiff images and run information from fio files.
        Applies the threshold and cutoff.

        -- run_nos : single run number or list of run numbers
        """
        
        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]
        run_nos = flatten(run_nos)

        for n in run_nos:
            if n not in self.runs.keys():
                self.runs[n] = None

        for run_no in run_nos:
            if self.runs[run_no] and self.runs[run_no]["complete"]:
                continue

            fio_file = "{0}_{1:05d}.fio".format(self.exp, run_no)
            fio_remote = os.path.join(self.datdir, fio_file)
            fio_local = os.path.join(self.localdir, fio_file)

            if not os.path.isfile(fio_local):
                a = load_fio(run_no, self.exp, self.datdir)
                if a and a["complete"]:
                    shutil.copyfile(fio_remote, fio_local)
            else:
                a = load_fio(run_no, self.exp, self.localdir)

            # default to remote folder if it exists
            if os.path.exists(self.datdir):
                img_root = self.datdir
            else:
                img_root = self.localdir
            img_folder = f"{self.exp}_{run_no:05d}"
            img_folder = os.path.join(img_root, img_folder, self.detector_type)

            # find image files and sort by collection time
            filepaths = glob(os.path.join(img_folder, "*.tiff"))
            filepaths = sorted(filepaths, key=os.path.getctime)
            filenames = [os.path.basename(f) for f in filepaths]

            img_list = []
            for i, f in enumerate(filenames):
                img = load_tiff(
                    f,
                    run_no,
                    self.exp,
                    self.datdir,
                    self.localdir,
                    self.bias_correct,
                    self.detector_type,
                )
                if img is None:
                    break
                img -= self.detfac
                bounds = (img > self.threshold) & (img < self.cutoff)
                img[~bounds] = 0
                img_list.append(img)
                sys.stdout.write(
                    "\r#{0:<4} {1:<3}/{2:>3} ".format(run_no, i+1, a["pnts"])
                )
            if not img_list:
                print(f"#{run_no:<4} -- no images loaded")
                continue
            if a["pnts"] != len(img_list):
                print("!!!")
            else:
                print()
            a["img"] = img_list
            self.runs[run_no] = a

    def transform(self, run_nos, ysca=1, fit=True):
        """ Transforms detector images into an array
        Applies defined ROI and stores summed intensity
        Stores the summed intensity along detector X (horiz) and Y (vert) axes

        -- run_nos : single run number or list of run numbers
        -- ysca : intensity scaling factor
        -- fit : if True, attempt to fit summed X and Y intensities
        """

        self.extract(run_nos)
        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        for run_no in run_nos:

            # sum up images if given a list of run_nos
            if isinstance(run_no, (list, tuple)):
                a = self.runs[run_no[0]]
                img = np.atleast_3d(np.array(a["img"])) / len(run_no)
                for r in run_no[1:]:
                    b = self.runs[r]
                    img = img + np.atleast_3d(np.array(b["img"])) / len(run_no)
            # single image
            else:
                a = self.runs[run_no]
                img = np.atleast_3d(np.array(a["img"]))
                img = img * ysca

            x = a["data"][a["auto"]]
            if len(x) != len(img):
                x = x[:img.shape[0]]

            y, roi, imgr = [], [], []
            rx, ry, imgx, imgy = [], [], [], []

            for im, xi in zip(img, x):
                try:
                    rc = self.roic(xi)  # roi centre defined by a function
                except TypeError:
                    rc = self.roic  # fixed value
                roiy = rc - (self.roih//2), rc+(self.roih//2)
                roi.append([self.roix[0], self.roix[1], roiy[0], roiy[1]])

                ri = im[roiy[0]:roiy[1], self.roix[0]:self.roix[1]]
                imgr.append(ri)
                y.append(np.nansum(ri))

                rx.append(np.arange(roiy[0], roiy[1]))
                ry.append(np.arange(self.roix[0], self.roix[1]))

                imgx.append(np.nansum(ri, axis=1))
                imgy.append(np.nansum(ri, axis=0))

            y = np.array(y)

            r1 = min(i[0] for i in roi)
            r2 = max(i[1] for i in roi)
            r3 = min(i[2] for i in roi)
            r4 = max(i[3] for i in roi)
            x_totx = np.arange(r3, r4)
            x_toty = np.arange(r1, r2)

            imgroi = np.atleast_3d(np.array(imgr))
            tot = np.nansum(imgroi, axis=0) / img.shape[0]
            totx = np.nansum(tot, axis=1)
            toty = np.nansum(tot, axis=0)

            f_txt = "fwhm: {0:.4f}\ncen: {1:.1f}\namp: {2:.1f}\nfra: {3:.1f}"
            if fit:
                try:
                    xfx, yfx, px = peak_fit(x_totx, totx)
                    xfy, yfy, py = peak_fit(x_toty, toty)
                    txtx = f_txt.format(px[1]*2, px[2], px[0], px[3])
                    txty = f_txt.format(py[1]*2, py[2], py[0], py[3])
                except:
                    fit = False

            a["x"], a["y"] = x, y
            a["roi"], a["imgr"] = roi, imgroi
            a["rx"], a["ry"], a["imgx"], a["imgy"] = rx, ry, imgx, imgy
            a["extent"] = [r1, r2, r3, r4]
            a["tot"], a["totx"], a["toty"] = tot, totx, toty
            a["x_totx"] = x_totx
            a["x_toty"] = x_toty
            a["fitted"] = fit
            if fit:
                a["xfx"], a["yfx"], a["px"], a["txtx"] = xfx, yfx, px, txtx
                a["xfy"], a["yfy"], a["py"], a["txty"] = xfy, yfy, py, txty

    def detector(self, run_no):
        """ plot raw features of the detector signal
        interactively step through individual images
        """

        self.extract(run_no)

        if "x" not in self.runs[run_no]:
            self.transform(run_no)
        a = self.runs[run_no]

        fig = plt.figure(figsize=(16.5, 8.5))
        gs0 = fig.add_gridspec(
            1, 2, left=0.05, right=0.99, top=0.99, bottom=0.04,
            width_ratios=[1, 1.35]
        )

        # averaged total
        gs1 = gs0[0].subgridspec(
            3, 2,  hspace=0.1, wspace=0.1,
            height_ratios=[1, 0.2, 0.4], width_ratios=[0.2, 1],
        )
        ax1 = fig.add_subplot(gs1[0, 1])
        ax2 = fig.add_subplot(gs1[0, 0])
        ax3 = fig.add_subplot(gs1[1, 1])
        ax4 = fig.add_subplot(gs1[2, :])

        fig.canvas.manager.set_window_title(f"#{run_no}")

        ax1.imshow(
            a["tot"],
            origin="lower",
            interpolation="hanning",
            vmin=self.vmin,
            vmax=self.vmax_average,
            cmap=plt.get_cmap("bone_r"),
            extent=a["extent"],
            aspect="auto"
        )

        ax2.plot(a["totx"], a["x_totx"], color="k", lw=0.75)
        ax3.plot(a["x_toty"], a["toty"], color="k", lw=0.75)

        if a["fitted"]:
            ax2.plot(a["yfx"], a["xfx"], color="k",lw=0.5)
            ax3.plot(a["xfy"], a["yfy"], color="k",lw=0.5)
            ax2.add_artist(AnchoredText(
                a["txtx"], 
                loc=1, 
                frameon=False, 
                pad=0.1, 
                prop=dict(fontsize="small")
            ))
            ax3.add_artist(AnchoredText(
                a["txty"],
                loc=2,
                frameon=False,
                pad=0.1,
                prop=dict(fontsize="small")
            ))

        r1, r2, r3, r4 = a["extent"]
        ax2.set_ylim(r3, r4)
        ax3.set_xlim(r1, r2)

        # interactive singles
        img, roi, x = a["img"], a["roi"], a["x"]
        imgx, imgy, rx, ry = a["imgx"], a["imgy"], a["rx"], a["ry"]

        gs2 = gs0[1].subgridspec(
            2, 2, hspace=0.05, wspace=0.05,
            height_ratios=[1, 0.2], width_ratios=[0.2, 1]
        )
        ax1 = fig.add_subplot(gs2[0,1])
        ax2 = fig.add_subplot(gs2[0,0])
        ax3 = fig.add_subplot(gs2[1,1])

        ax4.plot(a["x"], a["y"], color='k', lw=0.75)
        ax4.set_xlabel(a["auto"])
        ax4.set_ylabel("summed intensity")

        i = {"no":0}
        im = img[0]
        r = roi[0]

        cmap = copy.copy(plt.get_cmap("winter"))
        cmap.set_under("w")
        cmap.set_over("#F012BE")

        i["a"] = ax1.imshow(
            im,
            origin="lower",
            vmin=self.vmin,
            vmax=self.vmax_single,
            cmap=cmap,
            aspect="auto"
        )
        i["b"], = ax2.plot(imgx[0], rx[0], color="k", lw=0.75)
        i["c"], = ax3.plot(ry[0], imgy[0], color="k", lw=0.75)
        i["v"] = ax4.axvline(x[0], color="r", lw=0.5)
        i["t"] = ax4.text(0.05, 0.9, x[0], transform=ax4.transAxes)

        ax1.set_xlim(r[0], r[1])
        ax1.set_ylim(r[2], r[3])
        ax2.set_ylim(r[2], r[3])
        ax3.set_xlim(r[0], r[1])

        def do_plot(i):
            im = img[i["no"]]
            imx = imgx[i["no"]]
            imy = imgy[i["no"]]
            irx = rx[i["no"]]
            iry = ry[i["no"]]
            r = roi[i["no"]]
            i["a"].set_data(im)
            i["b"].set_xdata(imx)
            i["b"].set_ydata(irx)
            i["c"].set_xdata(iry)
            i["c"].set_ydata(imy)
            i["v"].set_xdata(x[i["no"]])
            i["t"].set_text(x[i["no"]])
            ax2.relim()
            ax3.relim()
            ax2.autoscale(axis="x")
            ax3.autoscale(axis="y")
            plt.draw()

        def press(event):
            if event.key == "left":
                if i["no"] > 0:
                    i["no"] -= 1
                    do_plot(i)
            elif event.key == "right":
                if i["no"] < img.shape[0]-1:
                    i["no"] += 1
                    do_plot(i)

        def click(event):
            if not event.dblclick and event.button == 1:
                if event.inaxes in [ax4]:
                    i["no"] = np.abs(x - event.xdata).argmin()
                    do_plot(i)

        fig.canvas.mpl_connect("button_press_event", click)
        fig.canvas.mpl_connect("key_press_event", press)

    def condition(
        self,
        run_nos,
        bins=None,
        oneshot_x=False,
        oneshot_y=False,
        oneshot_no=None,
        fit=False,
        xsca=1,
        x0=0
    ):
        """ Load, bin and save experiment run data to file.
        By default loads ROI intensity as a function of scanning motor

        -- run_nos : single run number or list of run numbers
        -- bins : bin size to rebin data
        -- oneshot_x : if True, load intensity summed along detector X axis
        -- oneshot_y : if True, load intensity summed along detector Y axis
        -- oneshot_no : Select single point/image in scan for oneshot summation
        -- fit : if True, attempt to fit a pseudo-voight profile to data
        -- xsca : x-axis scaling factor
        -- x0 : x-xaxis offset value  / x = (x - x0) * xsca
        """

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        self.transform(run_nos)

        for run_no in run_nos:
            if "x" not in self.runs[run_no]:
                self.transform(run_no)
            a = self.runs[run_no]

            if oneshot_x:
                if oneshot_no:
                    x = a["rx"][oneshot_no]
                    y = a["imgx"][oneshot_no]
                else:
                    x = a["x_totx"]
                    y = a["totx"]
                a["cond_type"] = "oneshot_x"
            elif oneshot_y:
                if oneshot_no:
                    x = a["ry"][oneshot_no]
                    y = a["imgy"][oneshot_no]
                else:
                    x = a["x_toty"]
                    y = a["toty"]
                a["cond_type"] = "oneshot_y"
            else:
                x, y = a["x"], a["y"]
                a["cond_type"] = a["auto"]

            if bins:
                x, y, _ = binning(x, y, bins)

            x = (x - x0) * xsca

        a["cond_x"], a["cond_y"] = x, y

        if fit:
            try:
                xf, yf, p = peak_fit(x, y)
                r = "#{0}  ".format(a["no"])
                r += "cen:{0:.4f}   ".format(p[2])
                r += "amp:{0:.2f}   ".format(p[0])
                r += "sig:{0:.4f}   ".format(p[1])
                r += "fwhm:{0:.4f}  ".format(p[1]*2)
                r += "fra:{0:.1f}   ".format(p[3])
                a["cond_xf"], a["cond_yf"] = xf, yf
                a["cond_p"], a["cond_r"] = p, r
                print(r)
            except:
                pass

        header = f"run: {a['no']}\nexp: {self.exp}\nbins: {bins}\n"
        if oneshot_x:
            header += "      y-pixel       intensity"
        elif oneshot_y:
            header += "      x-pixel       intensity"
        else:
            header += f"{a['auto']}    Intensity"
        np.savetxt(
            os.path.join(self.savedir, f"{self.exp}_{a['no']}_b{bins}.dat"),
            np.array([x, y]).T,
            fmt="% .8e",
            header=header
        )

    def plot(
        self,
        run_nos,
        ax=None,
        norm=False,
        fit=False,
        ystep=0,
        plot_trend=False,
        label_motor=None,
        **kwargs
    ):
        """ plot conditioned data """

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        if not ax:
            if plot_trend:
                fig = plt.figure(figsize=(7, 9), constrained_layout=True)
                gs = fig.add_gridspec(2, 3, height_ratios=[0.8, 1.0])
                axes = [
                    fig.add_subplot(gs[0,:]),
                    fig.add_subplot(gs[1,0]),
                    fig.add_subplot(gs[1,1]),
                    fig.add_subplot(gs[1,2])
                ]
            else:
                _, ax = plt.subplots(constrained_layout=True)

        if plot_trend:
            ax = axes[0]

        trend_x, trend_x0, trend_I, trend_fw = [], [], [], []
        for i, run_no in enumerate(run_nos):
            a = self.runs[run_no]
            x, y = a["cond_x"], a["cond_y"]
            if "cond_xf" in a and fit:
                xf, yf, p = a["cond_xf"], a["cond_yf"], a["cond_p"]
                if norm:
                    yf = yf/np.max(y)
            else:
                xf = None
            if norm:
                y = y/np.max(y)

            lab = f"#{run_no}"
            if label_motor:
                lab += f" {label_motor}: {a[label_motor]:.2f}"

            if fit and xf:
                trend_x.append(run_no)
                trend_x0.append(p[2])
                trend_fw.append(p[1]*2)
                trend_I.append(p[0])

            l, = ax.plot(x, y+i*ystep, label=lab, **kwargs)
            if fit and xf:
                ax.plot(xf, yf+i*ystep, color=l.get_color(), lw=0.5)

        ax.set_ylabel("Intensity")

        if a["cond_type"] == "oneshot_x":
            ax.set_xlabel("y-pixel")
        elif a["cond_type"] == "oneshot_y":
            ax.set_xlabel("x-pixel")
        else:
            ax.set_xlabel(a["cond_type"])

        if plot_trend:
            axes[1].plot(trend_x, trend_x0, ".-")
            axes[2].plot(trend_x, trend_fw, ".-")
            axes[3].plot(trend_x, trend_I, ".-")
            axes[1].set_ylabel("Center")
            axes[2].set_ylabel("FWHM")
            axes[3].set_ylabel("Amplitude")
            axes[3].set_ylim(0)
            for axi in axes:
                axi.set_xlabel("run number")

        ax.legend(fontsize="small")

    def track_signal(
        self,
        run_no,
        fit=True,
        maxsig=1200,
        detector_axis="x",
        title="no",
        plot=True,
        plot_trend=False,
        ystep=0
    ):
        """ Fit detector signal and follow as function of scanning motor """

        a = self.runs[run_no]

        rx, ry, imgx, imgy = a["rx"], a["ry"], a["imgx"], a["imgy"]

        if detector_axis == "x":  # horizontal
            r = rx
            im = imgx
        elif detector_axis == "y":  # vertical
            r = ry
            im = imgy

        txt_title = f"#{a['no']}"
        if title != "no":
            txt_title += f" {title}: {a[title]}"

        if plot:
            _, ax = plt.subplots(figsize=(6, 7), constrained_layout=True)
            ax.text(0.04, 0.98, txt_title, transform=ax.transAxes)

        trend_x0, trend_I, trend_fw = [], [], []
        for i, (x, y, xpos) in enumerate(zip(r, im, a["x"])):

            if plot:
                l, = ax.plot(x, y+i*ystep, lw=0.75)
                ax.text(x[0], y[0]+i*ystep, f"{xpos:.4f}", fontsize="small")

            if fit:
                try:
                    xf, yf, p = peak_fit(x, y)
                    if plot:
                        ax.plot(xf, yf+i*ystep, color=l.get_color(), lw=0.5)

                    if (p[1] < maxsig) and (p[2] > 0):
                        trend_I.append(p[0])
                        trend_x0.append(p[2])
                        trend_fw.append(p[1] * 2)
                    else:
                        trend_I.append(np.nan)
                        trend_x0.append(np.nan)
                        trend_fw.append(np.nan)    
                except:
                    trend_I.append(np.nan)
                    trend_x0.append(np.nan)
                    trend_fw.append(np.nan)
            else:
                com = np.sum(x * y)/np.sum(y)
                height = np.max(y) - np.min(y)
                half = x[np.abs(y - (height / 2 + np.min(y))).argmin()]
                wid = np.abs(half - com)*2
                trend_I.append(np.sum(y))
                trend_x0.append(com)
                trend_fw.append(wid)

        trend_x = a["x"]

        if plot_trend:
            _, ax = plt.subplots(
                1, 3, constrained_layout=True, figsize=(8, 4)
            )

            ax[0].plot(trend_x, trend_I)
            ax[1].plot(trend_x, trend_x0)
            ax[2].plot(trend_x, trend_fw)

            ax[0].set_ylabel("Intensity")
            ax[1].set_ylabel("Centre")
            ax[2].set_ylabel("FWHM")

            ax[0].text(0.05, 0.95, txt_title, transform=ax[0].transAxes)

        # print results to console
        results = np.array([
            range(len(trend_x)),
            trend_x,
            trend_I,
            trend_x0,
            trend_fw
        ]).T
        print(txt_title)
        print(tabulate(
            results,
            headers=["no", a["auto"], "intensity", "centre", "fwhm"]
        ))
        print()

        return results

    def logbook(
        self,
        nstart=None,
        nend=None,
        date=False,
    ):
        """
        display list of experiment runs
        -- nstart, nend : range of run numbers
        -- date : show date and time of run
        """

        if nstart:
            if nend is None:
                try:
                    latest = max(
                        iglob(os.path.join(self.datdir, "*.fio")),
                        key=os.path.getctime
                    )
                except ValueError:  # remote directory not present
                    print("Using Local Directory")
                    latest = max(
                        iglob(os.path.join(self.localdir, "*.fio")),
                        key=os.path.getctime,
                    )
                try:
                    latest = latest[:-4].split("_")[-1]
                    run_nos = range(nstart, int(latest) + 1)
                except ValueError:
                    return
            else:
                run_nos = range(nstart, nend + 1)

        for run_no in run_nos:
            a = load_fio(run_no, self.exp, self.datdir)
            if a is None:
                continue
            cmd_txt = " ".join(a["command"])
            out = "#{0:<4}{1}  ".format(run_no, cmd_txt)
            if date:
                out += a["date"]
            print(out)
        print()



import os
import numpy as np
import shutil

from skimage import io
from tifffile import TiffFileError
from numpy import sin, cos, sqrt, log, radians, arccos, pi
from scipy.optimize import curve_fit


def energy_to_wavelength(energy_in_eV):
    wavelength_in_angstrom = 12398.425 / energy_in_eV
    return wavelength_in_angstrom


def unit_cell_volume(a1, a2, a3, alpha1, alpha2, alpha3):
    """calculate unit cell volume (angles in radians)"""
    V = (
        a1 * a2 * a3 * sqrt(
            1 - cos(alpha1)**2 - cos(alpha2)**2 - cos(alpha3)**2 -
            2 * cos(alpha1) * cos(alpha2) * cos(alpha3)
        )
    )
    return V


def reciprocol_lattice(a1, a2, a3, alpha1, alpha2, alpha3):
    """calculate reciprocol lattice parameters (angles in radians)"""

    beta1 = arccos(
        (cos(alpha2) * cos(alpha3) - cos(alpha1)) / (sin(alpha2) * sin(alpha3))
    )

    beta2 = arccos(
        (cos(alpha1) * cos(alpha3) - cos(alpha2)) / (sin(alpha1) * sin(alpha3))
    )

    beta3 = arccos(
        (cos(alpha1) * cos(alpha2) - cos(alpha3)) / (sin(alpha1) * sin(alpha2))
    )

    V = unit_cell_volume(a1, a2, a3, alpha1, alpha2, alpha3)

    b1 = 2 * pi * a2 * a3 * sin(alpha1) / V
    b2 = 2 * pi * a1 * a3 * sin(alpha2) / V
    b3 = 2 * pi * a1 * a2 * sin(alpha3) / V

    return b1, b2, b3, beta1, beta2, beta3


def calc_dspacing(hkl, cell):
    """calculate dspacing from crystal lattice and orientation"""
    h, k, l = hkl
    a1, a2, a3, alpha1, alpha2, alpha3 = cell
    alpha1, alpha2, alpha3 = radians(alpha1), radians(alpha2), radians(alpha3)
    V = unit_cell_volume(a1, a2, a3, alpha1, alpha2, alpha3)

    S11 = a2**2 * a3**2 * sin(alpha1)**2
    S22 = a1**2 * a3**2 * sin(alpha2)**2
    S33 = a1**2 * a2**2 * sin(alpha3)**2
    S12 = a1 * a2 * a3**2 * (cos(alpha1) * cos(alpha2) - cos(alpha3))
    S23 = a2 * a3 * a1**2 * (cos(alpha2) * cos(alpha3) - cos(alpha1))
    S13 = a1 * a3 * a2**2 * (cos(alpha3) * cos(alpha1) - cos(alpha2))

    invD2 = (
        S11 * h ** 2
        + S22 * k ** 2
        + S33 * l ** 2
        + 2 * S12 * h * k
        + 2 * S23 * k * l
        + 2 * S13 * h * l
    )
    invD2 *= 1 / V ** 2
    d = 1 / sqrt(invD2)
    return d


def binning(x, y, n, photon_counting=False):
    """
    binning routine that also calculates error
    n: bin size
        - if integer, then simply divide array into steps of n
        - if float, then n corresponds to step size in eV
    photon_counting:
        - True: return events per bin
        - False: return intensity per bin averaged by counts
    """
    e = np.sqrt(y)
    if isinstance(n, int):  # strides
        xbin = int(len(x) / n)
    else:  # regularly spaced (in eV)
        xbin = np.arange(min(x), max(x), n)
    yi, xi = np.histogram(x, bins=xbin, weights=y)
    ei, _ = np.histogram(x, bins=xbin, weights=e ** 2)

    if not photon_counting:
        count, _ = np.histogram(x, bins=xbin)
        with np.errstate(divide="ignore", invalid="ignore"):
            yi = yi / count
            ei = np.sqrt(ei) / count

    xi = (xi[1:] + xi[:-1]) / 2
    return xi, yi, ei


def peak(x, a, sl, x0, f, bgnd):
    """basic pseudovoight profile with flat background"""
    m = np.full(len(x), bgnd)
    sg = sl / np.sqrt(2 * log(2))
    m += (
        (1 - f)
        * (a / (sg * np.sqrt(2.0 * np.pi)))
        * np.exp(-1.0 * (x - x0) ** 2.0 / (2.0 * sg ** 2.0))
    )
    m += f * (a / np.pi) * (sl / ((x - x0) ** 2.0 + sl ** 2.0))
    return m


def peak_fit(x, y):
    """
    peak fit routine that guesses initial values
    returns fitted peak xf,yf and parameters p = [amp, sig, cen, fra, bgnd]
    """
    fra = 0.5
    bgnd = float(np.min(y))
    height = np.max(y) - bgnd
    half = x[np.abs(y - height / 2 + bgnd).argmin()]
    cen = x[np.abs(y - np.max(y)).argmin()]
    fwhm = np.abs(half - cen) * 2
    sig = fwhm / 2
    amp = height * (sig * np.sqrt(2.0 * np.pi))

    xf = np.linspace(x.min(), x.max(), 1000)
    p0 = [amp, sig, cen, fra, bgnd]
    bounds = [(0, 1e-6, -1e9, 0, 0), (1e9, 1e6, 1e9, 1, 1e9)]
    p, _ = curve_fit(peak, x, y, p0, bounds=bounds)
    yi = peak(xf, *p)

    return xf, yi, p


def load_fio(run, exp, datdir):
    """
    .fio loader - returns everything as a dict
    data stored as a structured array with headers extracted from the fio
    contains a few helpers for P01 (qh, qk, ql, t_sample -> T)
    looks for '! Acquisition ended' to see if scan has been completed
    """
    a = {}
    head = []
    data = []
    complete = False
    fio_file = "{0}_{1:05d}.fio".format(exp, run)
    path = os.path.join(datdir, fio_file)
    if not os.path.isfile(path):
        print("#{0:<4} -- no .fio".format(run))
        return
    with open(path) as f:
        for line in f:
            l = line.strip()
            if l.startswith("%c"):
                command = next(f)[:-1].split()
                _, date = next(f).split(" started at ", maxsplit=1)
            elif line.startswith("%p"):
                break
        for line in f:
            if line.startswith("!"):
                break
            else:
                p, v = line.strip().split("=")
                try:
                    a[p.strip()] = float(v)
                except TypeError:
                    a[p.strip()] = v.strip()
        for line in f:
            l = line.strip().split()
            if not l:
                break
            if l[0] == "Col":
                head.append(l[2])
            else:
                try:
                    data.append([float(x) for x in l])
                except ValueError:
                    if line.startswith("! Acquisition ended"):
                        complete = True
    if head and data:

        data = [x for x in data if x]
        data = np.array(data)
        pnts = data.shape[0]
        data = data.view(dtype=[(n, float) for n in head])
        data = data.reshape(len(data))

        a["data"] = data
        a["auto"] = head[0]
        a["pnts"] = pnts
        a["th"] = a["rixs_th"]
        a["chi"] = a["rixs_chi"]

        if "q_h" in head:
            a["qh"] = np.average(data["q_h"])
            a["qk"] = np.average(data["q_k"])
            a["ql"] = np.average(data["q_l"])
        else:
            a["qh"], a["qk"], a["ql"] = 0.0, 0.0, 0.0

        if "t_coldhead" in head:
            a["t_coldhead"] = np.round(np.average(data["t_coldhead"]))
        if "t_sample" in head:
            a["t_sample"] = np.round(np.average(data["t_sample"]))
            a["T"] = a["t_sample"]
        else:
            a["T"] = 0.0

        a["command"] = command
        a["date"] = date.strip()
        a["time"] = float(command[-1])
        a["no"] = run
        a["complete"] = complete

        return a


def bias_correct_4output(rawimg):
    """
    correct bias of the Four DAC output mode of the greateyes detector ("EFGH")
    divide matrix up into four quadrants and subtract median from each quadrant
    """
    r, h = rawimg.shape[1], rawimg.shape[0]
    nrows,ncols = h//2, r//2
    img = (rawimg
           .reshape((h//nrows,nrows,-1,ncols))
           .swapaxes(1,2)
           .reshape((-1, nrows, ncols)))
    for i,im in enumerate(img):
        cen = int(np.median(im))
        img[i] -= cen
    img = np.block([[img[0], img[1]], [img[2], img[3]]])
    return img


def load_tiff(
    tiff,
    run,
    exp,
    datdir,
    localdir,
    bias_correct=False,
    detector="andor",
    ):
    """
    loads a tiff image and returns as a numpy array
    - if localdir defined, looks to local directory first,
      otherwise copies the remote file once loaded
    - catches corrupted tiff files from previous copy operation cancellation
    """

    # generate tiff filename from step number if given
    if isinstance(tiff, int):
        tiff = "{0}_{1:05d}_{2:04d}.tiff".format(exp, run, tiff)

    folder = "{0}_{1:05d}".format(exp, run)
    path_remote = os.path.join(datdir, folder, detector, tiff)

    if localdir:
        path_local = os.path.join(localdir, folder, detector, tiff)
        try:
            img = io.imread(path_local)
            # check if tiff file was corrupted on previous copy to local
            if img.shape[0] == 0:
                raise TiffFileError
        except (OSError, TiffFileError):
            try:
                os.makedirs(os.path.dirname(path_local), exist_ok=True)
                shutil.copyfile(path_remote, path_local)
                img = io.imread(path_local)
            except OSError:
                return
    else:
        try:
            img = io.imread(path_remote)
        except OSError:
            return

    if bias_correct:
        img = bias_correct_4output(img)

    return img


def flatten(*n):
    """flattens a lists of lists/ranges/tuples for loading"""
    return [
        e
        for a in n
        for e in (flatten(*a) if isinstance(a, (tuple, list, range)) else (a,))
    ]



#!/usr/bin/env python

"""P01-PLOT: quick plotting for experiments on P01 and P09 at DESY"""

__author__ = 'Joel Bertinshaw'
__credits__ = 'Martin Sundermann, Hlynur Gretarsson'
__copyright__ = 'Copyright 2021, Max Planck Institute for Solid State Research'
__license__ = 'GPL v3'

import os
import sys
import glob
import re
import tempfile
import subprocess
import platform

try:
    import fcntl

    nonblockread = True
except ImportError:
    nonblockread = False

import numpy as np

np.seterr(all="raise")

import matplotlib

matplotlib.use("Qt5Agg")
matplotlib.rcParams["xtick.top"] = True
matplotlib.rcParams["xtick.direction"] = "in"
matplotlib.rcParams["ytick.right"] = True
matplotlib.rcParams["ytick.direction"] = "in"
matplotlib.rcParams["xtick.labelsize"] = "small"
matplotlib.rcParams["ytick.labelsize"] = "small"
matplotlib.rcParams["axes.titlesize"] = "small"
matplotlib.rcParams["axes.labelsize"] = "small"
matplotlib.rcParams["axes.formatter.useoffset"] = False

from PyQt5 import QtCore, QtWidgets, QtGui, QtPrintSupport
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.widgets import Cursor, RectangleSelector
from matplotlib.offsetbox import AnchoredText
from os.path import basename, splitext
from scipy.optimize import least_squares
from datetime import datetime as dt
from copy import deepcopy

progname = "P01-PLOT"

REMOTE = False
FOLDER = None

HIDE = [
    "enc_dcm_ener",
    "epoch",
    "timestamp",
    "qh",
    "qk",
    "ql",
    "q_h",
    "q_k",
    "q_l",
    "e4cctrl_h/position",
    "e4cctrl_k/position",
    "e4cctrl_l/position",
    "h_position",
    "k_position",
    "l_position",
    "abs_attenfactor",
    "abs_position",
    "dettimesattenfactor_counts",
    "sumvfcs_counts",
]

TEMPCNTR = [
    "lks336_outputchannela",
    "lks336_outputchannelb",
    "lks336_outputchannelc",
    "lks336_outputchanneld",
    "t_coldhead",
    "t_sample",
]

TIMECNTR = [
    "exp_t01",
    "exp_t02",
    "eh1_t01",
    "eh1_t02",
    "oh2_t01",
    "oh2_t02",
    "eh2_t01",
    "eh2_t02",
    "eh3_t01",
    "eh3_t02",
]

ENERCNTR = ["enc_dcm_ener"]

HIDE += TIMECNTR

PRECISION = 4


def open_external(path, diff=False):
    if diff:
        paramdiff = fio_param_diff(path)
        path = os.path.join(tempfile.gettempdir(), "p01plotdiff.txt")
        with open(path, "w") as f:
            f.write(paramdiff)
    if platform.system() == "Darwin":
        subprocess.Popen(("open", path))
    elif platform.system() == "Windows":
        os.startfile(path)
    else:
        subprocess.Popen(["gedit", "-s", path])


def fio_param_diff(paths):
    with open(paths[0], "r") as f:
        al = f.readlines()
    with open(paths[1], "r") as f:
        bl = f.readlines()
    out = [[basename(p) for p in paths]]
    ml = len(out[0][0])
    for a, b in zip(al, bl):
        a, b = a.strip(), b.strip()
        if a.startswith("Col"):
            break
        elif a.startswith("user"):
            continue
        elif a != b or "scan" in a:
            out.append([a, b])
            if a.split()[0] in [
                "ascan",
                "a2scan",
                "a3scan",
                "dscan",
                "d2scan",
                "d3scan",
                "hklscan",
            ]:
                out.append(["-" * len(p) for p in [a, b]])
            if len(a) > ml:
                ml = len(a)
    out = ["{0}  |  {1}".format(l[0].ljust(ml), l[1]) for l in out]
    return "\n".join(out)


def find_nearest(array, value):
    return (np.abs(array - value)).argmin()


def get_numor(s):
    return int(re.findall(r"\d+", s)[-1])


def peak(x, amp, sig, cen, bgnd, frac):
    base = np.full_like(x, bgnd)
    sigg = sig / np.sqrt(2 * np.log(2))
    try:
        g = (
            (1 - frac)
            * (amp / (sigg * np.sqrt(2 * np.pi)))
            * np.exp(-1 * (x - cen) ** 2 / (2 * sigg ** 2))
        )
    except FloatingPointError:
        g = 0
    l = (frac) * (amp / np.pi) * (sig / ((x - cen) ** 2 + sig ** 2))
    return base + g + l


def peak_stats(x, y, deriv=False):
    try:
        centre = x[find_nearest(y, np.max(y))]
        msg = "max position: ( {0:.{2}f}, {1:.0f} )   ".format(
            centre, np.max(y), PRECISION
        )
        com = np.sum(x * y) / np.sum(y)
        msg += "com: {0:.{1}f}   ".format(com, PRECISION)
        if deriv:
            bgnd = 0
            miny, maxy = np.min(y), np.max(y)
            height = miny if abs(miny) > maxy else maxy
        else:
            bgnd = np.min(y)
            height = np.max(y) - bgnd
        half = x[find_nearest(y, height / 2 + bgnd)]
        wid = np.abs(half - com) * 2
        msg += "est. width: {0:.{1}f}".format(wid, PRECISION)
        return msg
    except (ValueError, FloatingPointError):
        return ""


def peak_fit(x, y, deriv=False):

    errfunc = lambda p, x, y: peak(x, *p) - y
    bounds = ([-np.inf, 0, -np.inf, -np.inf, 0], [np.inf, np.inf, np.inf, np.inf, 1])

    def init_params(x, y):
        bgnd = 0 if deriv else np.min(y)
        frac = 0.5
        height = np.max(y) - bgnd
        half = x[find_nearest(y, height / 2 + bgnd)]
        centre = x[find_nearest(y, np.max(y))]
        fwhm = np.abs(half - centre) * 2
        sigma = fwhm / 2
        amplitude = height * sigma * np.pi
        return [amplitude, sigma, centre, bgnd, frac]

    # least squares fit
    p0 = init_params(x, y)
    p = least_squares(errfunc, p0, args=(x, y), bounds=bounds)

    # try inverted peak
    p0inv = init_params(x, y * -1)
    pinv = least_squares(errfunc, p0inv, args=(x, y * -1), bounds=bounds)

    # compare residuals
    msg_pos = "top"
    if np.sum((y * -1 - peak(x, *pinv.x)) ** 2) < np.sum((y - peak(x, *p.x)) ** 2):
        p = pinv
        p.x[0] *= -1
        p.x[3] *= -1
        msg_pos = "bottom"

    msg = "amp: {0:.1f}\ncen: {1:.{3}f}\nfwhm: {2:.{3}f}".format(
        p.x[0], p.x[2], p.x[1] * 2, PRECISION
    )
    xi = np.linspace(min(x), max(x), 1000)
    yi = peak(xi, *p.x)

    return xi, yi, [msg, msg_pos]


class MplCanvas(FigureCanvas):
    def __init__(self, parent=None, width=5.8, height=4.8, dpi=100, replot=None):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        self.axes = self.fig.add_subplot(111)
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
        self.replot = replot
        self.zoomed_plot = False
        FigureCanvas.setSizePolicy(
            self, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        )
        FigureCanvas.updateGeometry(self)
        self.mpl_connect("key_press_event", self.press)
        self.mpl_connect("button_press_event", self.click)
        self.mpl_connect("button_release_event", self.release)
        self.cleanup()

    def reset(self):
        self.zoomed_plot = False
        self.axes.clear()

    def cleanup(
        self, plot_count=0, xlabel=None, ylabel=None, fitmsg=None, modemsg=None
    ):

        if xlabel:
            self.axes.set_xlabel(", ".join(sorted(set(xlabel))))
        if ylabel:
            ylabel = ", ".join(sorted(set(ylabel)))
            self.axes.set_ylabel(ylabel)
        if 0 < plot_count < 3:
            self.axes.legend(
                loc="lower right",
                fontsize="small",
                handlelength=1.5,
                ncol=2,
                labelspacing=0.3,
                handletextpad=0.5,
                borderaxespad=0.1,
                frameon=False,
                borderpad=0,
                bbox_to_anchor=(1, 1),
                bbox_transform=self.axes.transAxes,
            )
        elif 0 < plot_count < 21:
            self.axes.legend(
                loc=0,
                fontsize="small",
                handlelength=1.5,
                labelspacing=0.3,
                handletextpad=0.5,
            )

        if REMOTE:
            self.axes.format_coord = lambda x, y: ""
        else:
            self.cursor = Cursor(self.axes, lw=0.5, color="0.6", useblit=True)

        self.RS = RectangleSelector(
            self.axes,
            self.rect_callback,
            button=3,
            drawtype="box",
            useblit=True,
            rectprops=dict(fc="black", alpha=0.05),
        )
        self.RS.set_active(True)

        self.txt = AnchoredText(
            "",
            loc="upper left",
            bbox_to_anchor=(-0.12, -0.07),
            bbox_transform=self.axes.transAxes,
            frameon=False,
            borderpad=0,
            prop=dict(fontsize="small"),
        )
        self.axes.add_artist(self.txt)

        if modemsg:
            modemsg = ", ".join(modemsg)
            modemsg = AnchoredText(
                modemsg,
                loc="lower left",
                bbox_to_anchor=(0, 1),
                bbox_transform=self.axes.transAxes,
                frameon=False,
                borderpad=0,
                pad=0.1,
                prop=dict(fontsize="small"),
            )
            self.axes.add_artist(modemsg)

        if fitmsg:
            if fitmsg[1] == "bottom":
                loc = "lower left"
            elif fitmsg[1] == "top":
                loc = "upper left"
            fitmsg = AnchoredText(
                fitmsg[0], loc, prop=dict(fontsize="small"), frameon=False, pad=0.8
            )
            self.axes.add_artist(fitmsg)

        self.draw_idle()

    def press(self, event):
        if event.key == "l":
            if self.axes.get_yscale() == "linear":
                self.axes.set_yscale("log", nonposy="mask")
            else:
                self.axes.set_yscale("linear")
            self.draw_idle()

    def click(self, event):
        if not event.inaxes:
            return
        if event.button == 1:
            try:
                self.v.remove()
                self.h.remove()
            except:
                pass
            self.txt.txt.set_text(
                "( {0:.{2}f}, {1:.0f} )".format(event.xdata, event.ydata, PRECISION)
            )
            self.v = self.axes.axvline(event.xdata, color="k", lw=0.5)
            self.h = self.axes.axhline(event.ydata, color="k", lw=0.5)
            self.draw_idle()
        elif event.button == 3 and not REMOTE:
            self.cursor.disconnect_events()

    def release(self, event):
        if event.button == 3 and not REMOTE:
            self.cursor.connect_event("motion_notify_event", self.cursor.onmove)
            self.cursor.connect_event("draw_event", self.cursor.clear)

    def rect_callback(self, eclick, erelease):
        if eclick.xdata != erelease.xdata:
            xlim = sorted([eclick.xdata, erelease.xdata])
            ylim = sorted([eclick.ydata, erelease.ydata])
            if xlim[1] - xlim[0] > 1e-8 and ylim[1] - ylim[0] > 1e-8:
                self.axes.set_xlim(*xlim)
                self.axes.set_ylim(*ylim)
                self.zoomed_plot = True
        else:
            self.replot()
        self.draw_idle()


class ApplicationWindow(QtWidgets.QMainWindow):
    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setWindowTitle("application main window")

        self.exp_name = ""
        self.folder = FOLDER

        file_menu = QtWidgets.QMenu("File", self)
        file_menu.addAction("Filter experiments", self.select_exp_name)
        file_menu.addAction("Reload counters", self.reinit)
        file_menu.addSeparator()
        file_menu.addAction(
            "Save plot", self.save_on_click, QtCore.Qt.CTRL + QtCore.Qt.Key_S
        )
        file_menu.addAction(
            "Print plot", self.print_on_click, QtCore.Qt.CTRL + QtCore.Qt.Key_P
        )
        file_menu.addSeparator()
        file_menu.addAction("About", self.about)
        file_menu.addAction("Quit", self.fileQuit, QtCore.Qt.CTRL + QtCore.Qt.Key_Q)

        tool_menu = QtWidgets.QMenu("Options", self)
        self.autoCheck = QtWidgets.QAction("Follow experiment", self, checkable=True)
        self.motoCheck = QtWidgets.QAction(
            "Mark original motor position", self, checkable=True
        )
        self.timeCheck = QtWidgets.QAction("Scale by count time", self, checkable=True)
        self.motoCheck.triggered.connect(lambda: self.plot())
        self.timeCheck.triggered.connect(lambda: self.plot(dofit=False, deriv=False))
        tool_menu.addAction(self.autoCheck)
        tool_menu.addAction(self.motoCheck)
        tool_menu.addSeparator()
        tool_menu.addAction(self.timeCheck)

        self.menuBar().addMenu(file_menu)
        self.menuBar().addMenu(tool_menu)

        self.main_widget = QtWidgets.QWidget(self)
        l = QtWidgets.QHBoxLayout(self.main_widget)
        v = QtWidgets.QVBoxLayout()
        h1 = QtWidgets.QHBoxLayout()
        h2 = QtWidgets.QHBoxLayout()
        h3 = QtWidgets.QHBoxLayout()
        width = 202

        self.selectScans = QtWidgets.QListWidget()
        self.selectScans.setFixedWidth(width)
        self.selectScans.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.selectScans.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.selectScans.itemSelectionChanged.connect(
            lambda: self.plot(dofit=False, deriv=False, norm=False)
        )
        self.selectScans.customContextMenuRequested.connect(self.scansRightClick)

        self.selectCounters = QtWidgets.QListView()
        self.selectCounters.setFixedSize(width, 100)
        self.counters = QtGui.QStandardItemModel(self.selectCounters)
        self.counters.itemChanged.connect(
            lambda: self.plot(dofit=False, deriv=False, norm=False)
        )

        self.xCombo = QtWidgets.QComboBox(self)
        self.xCombo.addItem("Primary motor", "cstp")
        self.xCombo.addItem("Step number", "xstp")
        self.xCombo.addItem("Elapsed time", "tstp")
        self.xCombo.addItem("Incident energy", "estp")
        self.xCombo.activated.connect(self.pick_xaxis)

        self.plotButton = QtWidgets.QPushButton("Reset")
        self.plotButton.clicked.connect(
            lambda: self.plot(deriv=False, norm=False, dofit=False)
        )
        self.plotButton.setEnabled(False)

        self.deriButton = QtWidgets.QPushButton("Derivative")
        self.deriButton.clicked.connect(lambda: self.plot(deriv=True, norm=None))
        self.deriButton.setEnabled(False)

        self.normButton = QtWidgets.QPushButton("Minmax")
        self.normButton.clicked.connect(lambda: self.plot(norm=True, deriv=None))
        self.normButton.setEnabled(False)

        self.fitButton = QtWidgets.QPushButton("Fit")
        self.fitButton.clicked.connect(
            lambda: self.plot(dofit=True, norm=None, deriv=None)
        )
        self.fitButton.setEnabled(False)

        self.mplCanvas = MplCanvas(self.main_widget, replot=self.plot)
        self.mplCanvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.mplCanvas.setFocus()

        h1.addWidget(self.xCombo)
        h2.addWidget(self.normButton)
        h2.addWidget(self.deriButton)
        h3.addWidget(self.plotButton)
        h3.addWidget(self.fitButton)
        v.addWidget(self.selectScans)
        v.addWidget(self.selectCounters)
        v.addLayout(h1)
        v.addLayout(h2)
        v.addLayout(h3)
        l.addLayout(v)
        l.addWidget(self.mplCanvas)

        self.main_widget.setFocus()
        self.setCentralWidget(self.main_widget)

        self.status_message = QtWidgets.QLabel()
        self.statusBar().addWidget(self.status_message)

        self.initialise()

        def check_changes():
            try:
                self.plot(auto=True)
            finally:
                QtCore.QTimer.singleShot(2000, check_changes)

        QtCore.QTimer.singleShot(2000, check_changes)

    def initialise(self):
        self.fit_on = False
        self.deriv_on = False
        self.norm_on = False
        self.loaded = {}
        self.fio_counters = []
        self.flist = []
        self.ban_list = []
        self.pop_block = False
        self.plotted_scans = {}
        self.epoch_init = None
        self.xaxis = "cstp"

        self.populate()
        QtCore.QTimer.singleShot(500, lambda: self.selectScans.scrollToBottom())

        for counter in sorted(self.fio_counters):
            item = QtGui.QStandardItem(counter)
            item.setCheckable(True)
            self.counters.appendRow(item)
        self.selectCounters.setModel(self.counters)

        self.status_message.setText(" select a scan...")

    def populate(self):
        if self.pop_block:
            return
        self.pop_block = True
        fiotxt = "{}*.fio".format(self.exp_name)
        flist = glob.glob(os.path.join(self.folder, fiotxt))
        if nonblockread:
            flist.sort(key=lambda x: os.path.getmtime(x))
        else:
            flist.sort()

        if not flist:
            print("no .fio files found in {}\n".format(self.folder))
            print(HELP)
            sys.exit(2)

        for b in self.ban_list:
            try:
                flist.remove(b)
            except ValueError:
                pass
        pop_list = list(set(flist).difference(self.flist))
        if nonblockread:
            pop_list.sort(key=lambda x: os.path.getmtime(x))
        else:
            pop_list.sort()
        latest = flist[-1]
        item = False
        for i, f in enumerate(pop_list):
            n, a = self.load_fio(f)
            if not n:
                flist.remove(f)
                if f != latest:
                    self.ban_list.append(f)
                continue
            title = "{}  –  {}".format(n, a)
            item = QtWidgets.QListWidgetItem(title)
            item.setData(QtCore.Qt.UserRole, f)
            self.selectScans.addItem(item)
            if self.autoCheck.isChecked():
                if i == len(pop_list) - 1:
                    self.selectScans.selectionModel().clear()
                    self.selectScans.setCurrentItem(item)
        if item:
            rect = self.selectScans.viewport().contentsRect()
            bottom = self.selectScans.indexAt(rect.bottomLeft())
            if self.selectScans.item(bottom.row()) in [None, item]:
                self.selectScans.scrollToBottom()
        self.flist = flist
        self.pop_block = False

    def load_fio(self, path):
        head = []
        data = []
        complete = False
        with open(path) as f:

            if nonblockread:
                fd = f.fileno()
                flag = fcntl.fcntl(fd, fcntl.F_GETFL)
                fcntl.fcntl(fd, fcntl.F_SETFL, flag | os.O_NONBLOCK)
                flag = fcntl.fcntl(fd, fcntl.F_GETFL)

            filename = splitext(basename(path))[0]
            numor = get_numor(filename)
            a = {}
            for line in f:
                if line[:-1] == "%c":
                    scantype, _ = next(f).split(maxsplit=1)
                    _, date = next(f).split(" started at ", maxsplit=1)
                    try:
                        date = dt.strptime(date[:-1], "%a %b %d %H:%M:%S %Y")
                        date = date.timestamp()
                    except:
                        pass
                elif line.startswith("%p"):
                    break
            for line in f:
                if line[:-1] == "!":
                    break
                else:
                    p, v = line.split(" = ")
                    try:
                        a[p] = np.float(v[:-1])
                    except:
                        a[p] = v[:-1]
            for line in f:
                if line[:-1] == "%d":
                    break
            for line in f:
                if line.startswith("! Acquisition ended"):
                    complete = True
                    break
                elif not line[:-1]:
                    break
                l = line[:-1].split()
                if l[0] == "Col":
                    head.append(l[2])
                else:
                    data.append(l)

        if head and data:
            try:
                data = np.asarray(data, dtype=np.float)
            except ValueError:
                try:
                    data = np.asarray(data[:-1], dtype=np.float)
                except ValueError:
                    return False, False

            pnts = data.shape[0]
            data = data.view(dtype=[(n, np.float) for n in head])
            data = data.reshape(pnts)
            a["data"] = data
            a["auto"] = head[0]
            a["pnts"] = pnts
            a["scantype"] = scantype
            a["numor"] = numor
            a["path"] = path
            a["expname"] = filename[:-6]
            a["complete"] = complete
            a["date"] = date

            self.loaded[path] = a

            if scantype in ["a3scan", "d3scan", "hklscan"]:
                counters = [c for c in head[3:] if c not in HIDE]
                motors = ", ".join(head[:3]).replace("e6cctrl_", "q")
            elif scantype in ["a2scan", "d2scan"]:
                counters = [c for c in head[2:] if c not in HIDE]
                motors = ", ".join(head[:2]).replace("e6cctrl_", "q")
            else:
                counters = [c for c in head[1:] if c not in HIDE]
                motors = head[0].replace("e6cctrl_", "q")
            self.fio_counters = list(set(counters + self.fio_counters))
            return numor, motors

        return False, False

    def plot(self, dofit=None, auto=False, deriv=None, norm=None):

        self.populate()

        scans = []
        changed = False
        for s in self.selectScans.selectedItems():
            scans.append(s.data(QtCore.Qt.UserRole))
        for no in scans:
            if not self.loaded[no]["complete"]:
                self.load_fio(no)
                changed = True
        if auto:
            if changed:
                self.update_only(scans)
            return

        counter_count = 0
        for idx in range(self.counters.rowCount()):
            item = self.counters.item(idx)
            if item.checkState():
                counter_count += 1
        if not counter_count:
            self.mplCanvas.reset()
            self.mplCanvas.draw_idle()
            self.status_message.setText(" select a counter...")
            return

        if dofit is True:
            self.fit_on = True
        elif dofit is False:
            self.fit_on = False
        if deriv is True:
            self.deriv_on = True
        elif deriv is False:
            self.deriv_on = False
        if norm is True:
            self.norm_on = True
        elif norm is False:
            self.norm_on = False

        self.mplCanvas.reset()
        ax = self.mplCanvas.axes

        self.plotted_scans = {}
        self.epoch_init = None
        plot_count = 0
        counter_found = True
        fitmsg = None
        xlabel = []
        ylabel = []
        ctime = []
        for no in scans:
            l = self.loaded[no]
            motor = l["auto"]
            self.plotted_scans[no] = {}

            for idx in range(self.counters.rowCount()):
                item = self.counters.item(idx)
                counter = item.text()
                if counter in l["data"].dtype.names:
                    if item.checkState():

                        x, y, xi, yi, fitmsg, ps, tc = self.gen_xy(l, motor, counter)
                        ctime.append(tc)

                        if counter_count > 1:
                            lab = "#{} {}".format(l["numor"], counter)
                        else:
                            lab = "#{}".format(l["numor"])

                        (d,) = ax.plot(
                            x, y, ".-", mfc="w", lw=1, mew=1, ms=4, label=lab
                        )

                        if self.fit_on:
                            ax.plot(xi, yi, "--", lw=0.5, color=d.get_color())

                        if self.xaxis == "tstp":
                            xlabel.append("Elapsed Time (seconds)")
                        elif self.xaxis == "xstp":
                            xlabel.append("step")
                        elif self.xaxis == "estp":
                            xlabel.append("Incident Energy (eV)")
                        else:
                            xlabel.append(motor)
                        ylabel.append(counter)
                        self.plotted_scans[no][counter] = d
                        plot_count += 1

                elif item.checkState():
                    counter_found = False

        if plot_count == 0:
            if not scans:
                self.status_message.setText(" select a scan...")
            elif not counter_found:
                self.status_message.setText(" counter(s) not in .fio...")
        elif plot_count == 1:
            self.status_message.setText(" " + ps)
        else:
            self.status_message.setText("")

        if plot_count == 1:
            self.fitButton.setEnabled(True)
        else:
            self.fitButton.setEnabled(False)

        if plot_count > 0 and not self.deriv_on:
            self.deriButton.setEnabled(True)
        else:
            self.deriButton.setEnabled(False)

        if not self.norm_on:
            self.normButton.setEnabled(True)
        else:
            self.normButton.setEnabled(False)

        if self.fit_on or self.norm_on or self.deriv_on:
            self.plotButton.setEnabled(True)
        else:
            self.plotButton.setEnabled(False)

        if (
            self.motoCheck.isChecked()
            and plot_count > 0
            and len(scans) == 1
            and self.xaxis == "cstp"
            and motor in l
            and "_dmy" not in motor
        ):
            ax.axvline(l[motor], lw=0.5, color="0.7")

        ct = list(set(ctime))
        if len(ct) == 1 and not self.norm_on and not self.deriv_on:
            modemsg = ["{}s".format(ct[0])]
        else:
            modemsg = []
        if self.deriv_on:
            modemsg += ["Derivative"]
        if self.norm_on:
            modemsg += ["Normalised"]
        self.mplCanvas.cleanup(plot_count, xlabel, ylabel, fitmsg, modemsg)

    def update_only(self, scans):
        plot_count = 0
        for no in scans:
            l = self.loaded[no]
            motor = l["auto"]
            for idx in range(self.counters.rowCount()):
                item = self.counters.item(idx)
                counter = item.text()
                if counter in l["data"].dtype.names:
                    if item.checkState():
                        x, y, _, _, _, ps, _ = self.gen_xy(l, motor, counter)
                        self.plotted_scans[no][counter].set_data(x, y)
                        plot_count += 1
        if not plot_count:
            return
        if not self.mplCanvas.zoomed_plot:
            self.mplCanvas.axes.relim(visible_only=True)
            self.mplCanvas.axes.autoscale()
        self.mplCanvas.draw_idle()
        if plot_count == 1:
            self.status_message.setText(" " + ps)

    def gen_xy(self, l, motor, counter):
        y = deepcopy(l["data"][counter])
        if self.xaxis == "cstp":
            x = deepcopy(l["data"][motor])
        elif self.xaxis == "tstp":
            try:
                x = deepcopy(l["data"]["epoch"])
            except ValueError:
                x = deepcopy(l["data"]["timestamp"]) + l["date"]
            if not self.epoch_init:
                self.epoch_init = x[0]
            x -= self.epoch_init
        elif self.xaxis == "xstp":
            x = np.arange(len(y)) + 1
        elif self.xaxis == "estp":
            for EC in ENERCNTR:
                if EC in l["data"].dtype.names:
                    x = deepcopy(l["data"][EC])
        if self.deriv_on:
            y = y[np.argsort(x)]
            x = x[np.argsort(x)]
            y = np.diff(y)
            x = x[:-1]
        xi, yi, fitmsg = None, None, None
        if self.fit_on:
            try:
                xi, yi, fitmsg = peak_fit(x, y, self.deriv_on)
            except:
                self.fit_on = False
        if self.norm_on:
            try:
                if self.fit_on:
                    yi = (yi - min(y)) / (max(y) - min(y))
                y = (y - min(y)) / (max(y) - min(y))
            except:
                pass
        count_time = None
        for tc in TIMECNTR:
            if tc in l["data"].dtype.names:
                count_time = l["data"][tc][0]
                break
        if self.timeCheck.isChecked() and counter not in TEMPCNTR and not self.norm_on:
            if count_time:
                y /= count_time
                if self.fit_on:
                    yi /= count_time
                count_time = 1.0
            else:
                self.timeCheck.setChecked(False)
        ps = peak_stats(x, y, self.deriv_on)
        return x, y, xi, yi, fitmsg, ps, count_time

    def pick_xaxis(self):
        self.xaxis = self.xCombo.itemData(self.xCombo.currentIndex())
        self.plot(dofit=False, deriv=False, norm=False)

    def save_on_click(self):
        f, t = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save plot", "", "*.pdf;;*.png;;*"
        )
        if f:
            figname, figext = os.path.splitext(f)
            if not figext and t != "*":
                f = figname + t[1:]
            figsize = self.mplCanvas.fig.get_size_inches()
            self.mplCanvas.fig.set_size_inches(5.5, 4.5)
            self.mplCanvas.fig.savefig(f, dpi=150)
            self.mplCanvas.fig.set_size_inches(*figsize)
            self.mplCanvas.draw()

    def print_on_click(self):
        printer = QtPrintSupport.QPrinter(QtPrintSupport.QPrinter.HighResolution)
        dialog = QtPrintSupport.QPrintDialog(printer, self)
        if dialog.exec_() == QtPrintSupport.QPrintDialog.Accepted:
            painter = QtGui.QPainter(printer)
            figfile = os.path.join(tempfile.gettempdir(), "p01plot.png")
            figsize = self.mplCanvas.fig.get_size_inches()
            self.mplCanvas.fig.set_size_inches(5.5, 4.5)
            self.mplCanvas.fig.savefig(figfile, dpi=900)
            self.mplCanvas.fig.set_size_inches(*figsize)
            self.mplCanvas.draw()
            image = QtGui.QImage(figfile)
            painter.drawImage(0, 0, image)
            painter.end()

    def scansRightClick(self, QPos):
        self.listMenu = QtWidgets.QMenu()
        menu_openfio = self.listMenu.addAction("Open .fio")
        menu_diffcheck = self.listMenu.addAction("Param. diff")
        if len(self.selectScans.selectedItems()) == 0:
            menu_openfio.setDisabled(True)
        if len(self.selectScans.selectedItems()) != 2:
            menu_diffcheck.setDisabled(True)
        menu_openfio.triggered.connect(self.openFioClicked)
        menu_diffcheck.triggered.connect(self.diffcheckClicked)
        parentPosition = self.selectScans.mapToGlobal(QtCore.QPoint(0, 0))
        self.listMenu.move(parentPosition + QPos)
        self.listMenu.show()

    def openFioClicked(self):
        for s in self.selectScans.selectedItems():
            no = s.data(QtCore.Qt.UserRole)
            open_external(self.loaded[no]["path"])

    def diffcheckClicked(self):
        fiofiles = []
        for s in self.selectScans.selectedItems():
            no = s.data(QtCore.Qt.UserRole)
            fiofiles.append(self.loaded[no]["path"])
        open_external(sorted(fiofiles), diff=True)

    def select_exp_name(self):
        self.pop_block = True
        title = "Filter Experiments"
        header = "Select experiment (filename prefix)" + " " * 25

        flist = glob.glob(os.path.join(self.folder, "*.fio"))
        flist = [os.path.basename(f).rsplit("_", 1)[0] for f in flist]
        flist = [""] + list(set(flist))
        idx = flist.index(self.exp_name)
        flist[0] = "All experiments"
        txt, ok = QtWidgets.QInputDialog.getItem(self, title, header, flist, idx, False)
        self.pop_block = False
        if ok:
            self.exp_name = txt if txt != "All experiments" else ""
            self.reinit()

    def reinit(self):
        self.counters.removeRows(0, self.counters.rowCount())
        self.selectScans.clear()
        self.initialise()

    def fileQuit(self):
        self.close()

    def about(self):
        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(QtCore.QByteArray.fromBase64(ICON))
        msg = QtWidgets.QMessageBox(self)
        msg.setWindowIcon(QtGui.QIcon(pixmap))
        msg.setIconPixmap(pixmap)
        msg.setText(ABOUTMSG)
        msg.setWindowTitle("About")
        msg.setStandardButtons(QtWidgets.QMessageBox.Ok)
        msg.show()


HELP = """
p01plot [directory] [--remote -r] [--help -h]
directory : location to look for .fio data files
            defaults to /gpfs/current/raw, then current directory
--remote : remove cursor to speed up remote connections
--help : show this menu
"""

ABOUTMSG = """<h3>{0}</h3>
<p><i>Quick plotting for experiments on P01 (and P09) at DESY.</i></p>
<p>Created by {1}<br>Thanks to {2}</p><p>{3}</p><p><pre>{4}</pre></p>
""".format(progname, __author__, __credits__, __copyright__, HELP)


ICON = b" "


def main():

    global REMOTE
    global FOLDER

    for arg in sys.argv:
        if arg in ["--help", "-h"]:
            print(HELP)
            sys.exit()
        if arg in ["--remote", "-r"]:
            REMOTE = True
    for arg in ["--remote", "-r", "--help", "-h"]:
        if arg in sys.argv:
            sys.argv.remove(arg)

    try:
        FOLDER = os.path.abspath(sys.argv[1])
        if not os.path.isdir(FOLDER):
            print("cannot find {}\n".format(FOLDER))
            print(HELP)
            sys.exit()
    except IndexError:
        if os.path.isdir("/gpfs/current/raw"):
            FOLDER = "/gpfs/current/raw"
        else:
            FOLDER = os.getcwd()

    qApp = QtWidgets.QApplication(sys.argv)
    aw = ApplicationWindow()
    aw.setWindowTitle(progname + " - " + FOLDER)

    pixmap = QtGui.QPixmap()
    pixmap.loadFromData(QtCore.QByteArray.fromBase64(ICON))
    aw.setWindowIcon(QtGui.QIcon(pixmap))

    aw.show()
    sys.exit(qApp.exec_())


if __name__ == "__main__":
    main()



""" reduction class for rowland circle spectrometer """

import os
import sys
import shutil
import numpy as np
import matplotlib.pyplot as plt
import scipy.ndimage

from numpy import pi, cos, tan, arcsin
from matplotlib.patches import Rectangle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from copy import deepcopy
from glob import iglob

from .tools import load_fio, load_tiff
from .tools import calc_dspacing, binning, peak_fit, flatten

PIX_EN_CONV = 13.5e-6  # andor detector pixel size
SR_LIMIT = 50  # minimum ring current in mA to identify beam dump

plt.rcParams["xtick.top"] = True
plt.rcParams["ytick.right"] = True
plt.rcParams["font.size"] = 8
plt.rcParams["axes.titlesize"] = "medium"
plt.rcParams["figure.titlesize"] = "medium"


def pix_to_E(energy, dspacing):
    """convert y-axis pixel position to energy dispersion from analyser"""
    wl = 12398.4193 / energy
    th = arcsin(wl / (2 * dspacing))
    l = 2 * cos(pi / 2 - th)
    dE = energy * PIX_EN_CONV / (l * tan(th))
    return dE


class irixs:
    def __init__(
        self,
        exp,
        y0=None,
        y0_fallback=None,
        roix=[0, 2048],
        roiy=[0, 2048],
        roih=None,
        threshold=1010,
        cutoff=1800,
        detfac=935,
        photon_factor=750,
        E0=None,
        analyser=None,
        photon_counting=False,
        photon_event_threshold=400,
        photon_max_events=0,
        datdir_remote="/gpfs/current/raw",
        datdir_local="raw"
    ):
        """
        exp -- experiment filename prefix

        y0 -- corresponding vertical pixel position on detector
        - can be given as single value or as a dict for each run
        y0_fallback -- if a dict is used for y0, use y0_fallback for runs not included the dict
        E0 -- elastic energy: typically use None to extract from .fio file

        roix -- detector region of interest
        roiy -- detector region of interest
        roih -- define height of roi instead of roiy, using y0 as centre (preferred)

        threshold -- minimum to kill readout noise (use histogram to refine)
        cutoff -- upper limit to kill cosmic rays (use histogram to refine)
        detfac -- andor detector factor (0 if andor bgnd sub is enabled, 935 otherwise)

        analyser -- analyser crystal lattice and reflection.
        - None defaults to quartz (1,0,2)
        photon_factor -- conversion between detector intensitiy and photon count
        - should be = 788, seems closer to 750
        - used to calculate correct intensities and systematic error

        photon_counting -- enable photon counting mode
        photon_event_threshold -- threshold intensity for a contigous detector event
        photon_max_events -- maximum multiple events to correct for (0 to disable correction)
        - photon_counting only works if the count rate on the detector is low
        """

        self.exp = exp
        self.runs = {}

        self.E0 = E0
        self.y0 = y0
        self.roix = roix
        self.roiy = roiy
        self.roih = roih

        self.threshold = threshold
        self.cutoff = cutoff
        self.detfac = detfac
        self.photon_factor = photon_factor

        self.photon_counting = photon_counting
        self.event_min = photon_event_threshold
        self.max_events = photon_max_events

        quartz = [(1, 0, 2), (4.9133, 4.9133, 5.4053, 90, 90, 120)]
        self.analyser = quartz if analyser is None else analyser
        self.dspacing = calc_dspacing(self.analyser[0], self.analyser[1])

        self.datdir = datdir_remote
        self.localdir = datdir_local
        if self.localdir:
            os.makedirs(self.localdir, exist_ok=True)

        self.savedir_dat = "dat"
        self.savedir_con = "con"
        self.savedir_det = "det"
        self.savedir_fig = "fig"
        os.makedirs(self.savedir_dat, exist_ok=True)
        os.makedirs(self.savedir_con, exist_ok=True)
        os.makedirs(self.savedir_det, exist_ok=True)
        os.makedirs(self.savedir_fig, exist_ok=True)

        self.corr_shift = False  # distortion correction

    def load(self, run_nos, load_images=True):
        """
        load data & parameters from fio and tiff files
        - stores runs as dicts in the self.runs dict
        - only copies to local directory when a run is completed
        - .fio files loaded first
        - performs tiff thresholding and cutoff (and stores values used)
        - stores parameters to be used for data conditioning later
        - should be smart enough to only reload/recondition runs when necessary

        runs -- numbers of runs, can be given as single run or list or list of lists
        load_images -- only load image data if True
        """

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]
        run_nos = flatten(run_nos)

        for n in run_nos:
            if n not in self.runs.keys():
                self.runs[n] = None

        for n in run_nos:

            if self.runs[n] and self.runs[n]["complete"]:
                continue

            # extract metadata
            fio_path = "{0}/{1}_{2:05d}.fio"
            path_remote = fio_path.format(self.datdir, self.exp, n)
            if self.localdir:
                path_local = fio_path.format(self.localdir, self.exp, n)
                if not os.path.isfile(path_local):
                    a = load_fio(n, self.exp, self.datdir)
                    if a and a["complete"]:
                        os.makedirs(self.localdir, exist_ok=True)
                        shutil.copyfile(path_remote, path_local)
                else:
                    a = load_fio(n, self.exp, self.localdir)
            else:
                a = load_fio(n, self.exp, self.datdir)
            if a is None:
                continue

            # define incoming and outgoing energies
            a["EI"] = a["dcm_ener"]
            if a["auto"] == "exp_dmy01":
                a["EF"] = np.full(a["pnts"], a["rixs_ener"])
            else:
                a["EF"] = a["data"]["rixs_ener"]

            self.runs[n] = a

        if not load_images:
            return

        to = self.threshold - self.detfac
        co = self.cutoff - self.detfac

        for n in run_nos:

            a = self.runs[n]
            if not a:
                continue

            if isinstance(self.y0, dict):
                try:
                    y0 = self.y0[n]
                except KeyError:
                    try:
                        y0 = self.y0_fallback
                    except AttributeError:
                        print("#{0:<4} -- y0 not given".format(n))
                        continue
            else:
                y0 = self.y0
            if self.roih:
                try:
                    roiy = [y0 + self.roih[0], y0 + self.roih[1]]
                except IndexError:
                    roiy = [y0 - self.roih // 2, y0 + self.roih // 2]
            else:
                roiy = self.roiy

            a["roix"], a["roiy"], a["y0"] = self.roix, roiy, y0

            if "to" in a and to == a["to"] and co == a["co"] and a["complete"]:
                continue

            if "img" not in a or a["img"] is None:
                imtest = load_tiff(0, n, self.exp, self.datdir, self.localdir)
                if imtest is None:
                    print("#{0:<4} -- no images".format(n))
                    a["img"] = None
                    continue
                else:
                    a["img"] = []

            for i, _ in enumerate(a["EF"]):
                if i > len(a["img"]) - 1:
                    img = load_tiff(i, n, self.exp, self.datdir, self.localdir)
                    if img is None:
                        print("!!!")
                        break
                    if img is not None:
                        img -= self.detfac
                        img[~np.logical_and(img > to, img < co)] = 0
                        a["img"].append(img)
                    sys.stdout.write(
                        "\r#{0:<4} {1:<3}/{2:>3} ".format(n, i + 1, a["pnts"])
                    )
                    if i + 1 == a["pnts"]:
                        sys.stdout.write("\n")
                    sys.stdout.flush()

            a["threshold"] = self.threshold
            a["cutoff"] = self.cutoff
            a["detfac"] = self.detfac
            a["to"], a["co"] = to, co

    def logbook(
        self,
        nstart=None,
        nend=None,
        run_nos=None,
        extras=["th"],
        hkl=False,
        date=False,
        only_rixs=True,
    ):
        """
        tabulate runs
        - shows if ring current was too low (SR_LIMIT)
        - shows if rixs_ener is different to dcm_ener

        nstart,nend -- range of runs
        runs -- specific runs
        extras,hkl,date -- extra parameters to display
        only_rixs -- only tabulate rixs runs
        """
        if not nstart and not run_nos:
            run_nos = list(self.runs.keys())

        if nstart:
            if nend is None:
                try:
                    latest = max(
                        iglob(os.path.join(self.datdir, "*.fio")), key=os.path.getctime
                    )
                except ValueError:
                    print("Using Local Directory")
                    latest = max(
                        iglob(os.path.join(self.localdir, "*.fio")),
                        key=os.path.getctime,
                    )
                try:
                    latest = latest[:-4].split("_")[-1]
                    run_nos = range(nstart, int(latest) + 1)
                except ValueError:
                    return
            else:
                run_nos = range(nstart, nend + 1)

        self.load(run_nos, False)
        for run_no in run_nos:
            out = ""
            a = self.runs[run_no]
            if a is None:
                continue
            try:
                command = a["command"]
                scantype, motor = command[:2]
                m1, m2, pnt, t = [float(c) for c in command[2:]]
            except IndexError:
                continue
            if np.any(a["data"]["sr_current"] < SR_LIMIT):
                dump = "*"
            else:
                dump = " "
            out += "#{0:<4}{1}{2:>13} ".format(run_no, dump, motor)
            if motor in ["rixs_ener"]:
                out += "{0:>+6.2f} > {1:+4.2f}".format(m1, m2)
            else:
                if only_rixs:
                    continue
                out += " {0:<13}".format("")
            out += " {0:3.0f}pnt {1:4.0f}s  ".format(pnt, t)
            out += "{0:7.1f}eV".format(a["dcm_ener"])
            if a["dcm_ener"] != a["rixs_ener"]:
                out += "* "
            else:
                out += "  "
            out += "{0:3.0f}K  ".format(a["T"])
            if hkl:
                qh = np.round(a["qh"], 2) + 0
                qk = np.round(a["qk"], 2) + 0
                ql = np.round(a["ql"], 2) + 0
                if not qh and not qk and not ql:
                    out += " " * 18
                else:
                    out += "({0: 3.1f} {1: 3.1f} {2:4.1f})  ".format(qh, qk, ql)
            for ex in extras:
                out += "{0}:{1:6.2f}  ".format(ex, a[ex])
            if date:
                out += " " + a["date"]
            print(out)
        print()

    def detector(
        self,
        run_nos,
        oneshot=None,
        fit=False,
        plot=True,
        com=False,
        vmax=10,
        savefig=False,
        use_distortion_corr=True,
    ):
        """
        data reduction without energy binning
        - if oneshot = False, returns summed intensity of ROI window vs step
        - if oneshot = True, returns total summed and flattened counts vs detector x-axis

        runs -- numbers of runs
        oneshot -- sum detector mode
        fit -- try and fit a peak
        plot -- plots all detector images summed together, as well as conditioned data
        com -- also plot (and store) center of mass along x- and y- detector axes
        vmax -- colormap threshold (10 is pretty good in most cases)
        """
        roic = "#F012BE"
        self.load(run_nos)

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        for run_no in run_nos:
            a = self.runs[run_no]
            if a is None or a["img"] is None:
                continue

            savefile = "{0}/{1}_{2:05d}_det.txt".format(
                self.savedir_det, self.exp, run_no
            )

            step, roix, roiy, y0 = a["auto"], a["roix"], a["roiy"], a["y0"]

            if oneshot is None:
                if step == "exp_dmy01":
                    oneshot = True
                else:
                    oneshot = False

            x = []
            y = []
            comV = []
            comH = []

            imgarr = np.atleast_3d(np.array(a["img"]))
            imtotal = np.nansum(imgarr, axis=0) / imgarr.shape[0]
            imgarr = imgarr[:, roiy[0] : roiy[1], roix[0] : roix[1]]

            if use_distortion_corr and self.corr_shift is not False:
                for sh, (c1, c2) in zip(self.corr_shift, self.corr_regions):
                    c1, c2 = c1 + roix[0], c2 + roix[0]
                    imtotal[:, c1:c2] = np.roll(imtotal[:, c1:c2], sh, axis=0)

            if oneshot:
                x = np.arange(roiy[0], roiy[1])
                y = np.nansum(imgarr, axis=(0, 2)) / imgarr.shape[0]
            else:
                for i, ef in enumerate(a["EF"]):
                    try:
                        img = imgarr[i]
                    except IndexError:
                        continue
                    yi = np.nansum(img)
                    xi = ef
                    x.append(xi)
                    y.append(yi)
                    if com:
                        rx = range(roiy[0], roiy[1])
                        ry = range(roix[0], roix[1])
                        cv = np.nansum(rx * np.nansum(img, axis=1)) / yi
                        ch = np.nansum(ry * np.nansum(img, axis=0)) / yi
                        comV.append(cv)
                        comH.append(ch)
                x, y = np.array(x), np.array(y)

            a["x"], a["y"], a["e"] = x, y, False
            a["label"] = run_no

            header = "experiment: {0}\n".format(self.exp)
            header += "run: {0}\n".format(run_no)
            header += "date: {}\n".format(a["date"])
            header += "command: {0}\n".format(" ".join(a["command"]))
            header += "dcm_ener: {0}\n".format(a["dcm_ener"])
            header += "rixs_ener: {0}\n".format(a["rixs_ener"])
            header += "det_threshold: {0}\n".format(a["threshold"])
            header += "det_cutoff: {0}\n".format(a["cutoff"])
            header += "det_factor: {0}\n".format(a["detfac"])
            header += "det_roix: {0}\n".format(roix)
            header += "det_roiy: {0}\n\n".format(roiy)

            if oneshot:
                header += "{0:>24}{1:>24}".format("y-pixel", "counts")
                save_array = np.array([x, y]).T
            elif com:
                header += "{0:>24}{1:>24}{2:>24}{3:>24}".format(
                    step, "roi-counts", "vert-COM", "horiz-COM"
                )
                comV, comH = np.array(comV), np.array(comH)
                a["comV"], a["comH"] = comV, comH
                save_array = np.array([x, y, comV, comH]).T
            else:
                header += "{0:>24}{1:>24}".format(step, "roi-counts")
                save_array = np.array([x, y]).T
            np.savetxt(savefile, save_array, header=header)

            if fit:
                a["xf"], a["yf"], a["p"] = peak_fit(x, y)
                report = "#{0:<4} (det)  ".format(run_no)
                report += "cen:{0:.4f}   ".format(a["p"][2])
                report += "amp:{0:.2f}   ".format(a["p"][0])
                report += "fwhm:{0:.3f}   ".format(a["p"][1] * 2)
                report += "fra:{0:.1f}   ".format(a["p"][3])
                report += "bg:{0:.3f}".format(a["p"][3])
                print(report)
            else:
                a["p"] = False

            a["oneshot"] = oneshot
            a.pop("E0", None)

            if plot:
                if com:
                    fig, ax = plt.subplots(
                        2, 2, constrained_layout=True, figsize=(8.5, 8)
                    )
                    ax = ax.flatten()
                else:
                    fig, ax = plt.subplots(1, 2, figsize=(8.5, 4))
                    fig.subplots_adjust(0.06, 0.15, 0.98, 0.93)

                plt.suptitle(
                    "#{}".format(a["label"]), ha="left", va="top", x=0.005, y=0.995
                )

                im = ax[0].imshow(
                    imtotal,
                    origin="lower",
                    vmax=vmax,
                    cmap=plt.get_cmap("bone_r"),
                    interpolation="hanning",
                )

                rect = Rectangle(
                    (roix[0], roiy[0]),  # xy origin, width, height
                    roix[1] - roix[0],
                    roiy[1] - roiy[0],
                    linewidth=0.5,
                    linestyle="dashed",
                    edgecolor=roic,
                    fill=False,
                )
                ax[0].add_patch(rect)
                ax[0].axhline(y0, color=roic, lw=0.5, dashes=(2, 2))

                div = make_axes_locatable(ax[0])
                cax = div.append_axes("right", size="4%", pad=0.1)
                fig.colorbar(im, cax=cax)

                ax[0].set_title("Summed Detector Map")
                ax[0].set_xlabel("x-pixel")
                ax[0].set_ylabel("y-pixel")
                ax[0].tick_params(which="both", direction="out", length=2)
                ax[0].xaxis.set_major_locator(plt.MultipleLocator(400))
                ax[0].yaxis.set_major_locator(plt.MultipleLocator(400))
                ax[0].xaxis.set_minor_locator(plt.MultipleLocator(100))
                ax[0].yaxis.set_minor_locator(plt.MultipleLocator(100))

                ax[1].plot(x, y, lw=1, color="#001F3F")
                if oneshot:
                    ax[1].set_xlabel("y-pixel")
                    ax[1].set_title("Integrated")
                else:
                    ax[1].ticklabel_format(axis="y", style="sci", scilimits=(0, 0))
                    ax[1].set_title("Counts in ROI")

                if a["p"] is not False:
                    ax[1].plot(a["xf"], a["yf"], color="#001F3F", dashes=(2, 8), lw=0.5)
                    fr = "fwhm: {:.3f}\ncen: {:.2f}".format(a["p"][1] * 2, a["p"][2])
                    ax[1].text(
                        0.025,
                        0.975,
                        fr,
                        va="top",
                        transform=ax[1].transAxes,
                        fontsize="small",
                        linespacing=1.3,
                    )

                if com:
                    ax[2].plot(a["x"], a["comH"], lw=1, color="#0074D9")
                    ax[2].set_title("Horizontal COM")
                    ax[2].set_ylabel("x-pixel")
                    ax[2].set_ylim(roix[0], roix[1])

                    ax[3].plot(a["x"], a["comV"], lw=1, color="#FF4136")
                    ax[3].set_title("Vertical COM")
                    ax[3].set_ylabel("y-pixel")
                    ax[3].set_ylim(roiy[0], roiy[1])

                for axi in ax[1:]:
                    axi.minorticks_on()
                    if not oneshot:
                        axi.set_xlabel(step)
                        for l in axi.get_xmajorticklabels():
                            l.set_rotation(30)

                if savefig:
                    if savefig is True:
                        savename = "{0}/det_s{1}_{2}.pdf".format(
                            self.savedir_fig, a["label"], a["auto"]
                        )
                    else:
                        savename = "{0}/{1}_s{2}.pdf".format(
                            self.savedir_fig, savefig, a["label"]
                        )
                    plt.savefig(savename, dpi=300)

    def condition(
        self, bins, run_nos, fit=False, use_distortion_corr=True, drop_beamdump=False
    ):
        """
        main data reduction routine, returns signal vs energy
        - stores result in self.runs[run]['x'] and self.runs[run]['y']
        - elastic line set using y0 and E0 (E0 normally set by hrm_ener)
        - saves binned and unbinned datasets to file
        - reloads data before starting condition (which also update parameters if changed)

        bins -- 0 to disable binning, set in eV (float) or in array stride (integer)
        runs -- numbers of runs
        - single value or list of runs
        - list of list will stitch runs under the first run number
        fit -- fit peak
        use_distortion_correction -- run calc_distortion to determine correction first
        drop_beamdump -- ignore runs measured during a beamdump
        """
        self.load(run_nos)
        if isinstance(run_nos, int):
            run_nos = [run_nos]

        for run_no in run_nos:

            if isinstance(run_no, int):
                run_no = [run_no]

            x, y, ns = [], [], []
            for n in run_no:
                a = self.runs[n]
                if a is None or "img" not in a or a["img"] is None:
                    continue
                if a["auto"] not in ["rixs_ener", "dcm_ener", "exp_dmy01"]:
                    continue
                ns.append(n)

                roix, roiy, y0 = a["roix"], a["roiy"], a["y0"]
                xinit = np.arange(roiy[0], roiy[1])
                if self.photon_counting:
                    xinit = np.tile(xinit, (roix[1] - roix[0], 1)).T

                for ef, img, sr in zip(a["EF"], a["img"], a["data"]["sr_current"]):

                    if drop_beamdump and sr < SR_LIMIT:
                        continue

                    img = deepcopy(img[:, roix[0] : roix[1]])

                    if use_distortion_corr and self.corr_shift is not False:
                        for sh, (c1, c2) in zip(self.corr_shift, self.corr_regions):
                            img[:, c1:c2] = np.roll(img[:, c1:c2], sh, axis=0)

                    img = img[roiy[0] : roiy[1]]
                    if self.photon_counting:
                        lbl, nlbl = scipy.ndimage.label(img)
                        try:
                            yi = scipy.ndimage.labeled_comprehension(
                                img, lbl, range(1, nlbl + 1), np.sum, float, 0
                            )
                            xi = scipy.ndimage.labeled_comprehension(
                                xinit, lbl, range(1, nlbl + 1), np.mean, float, 0
                            )
                            xi = (xi - y0) * pix_to_E(ef, self.dspacing) + ef
                            xi = xi[yi > self.event_min]
                            yi = yi[yi > self.event_min]
                        except ValueError:
                            pass
                    else:
                        yi = np.sum(img, axis=1)
                        xi = (xinit - y0) * pix_to_E(ef, self.dspacing) + ef
                    x.extend(xi)
                    y.extend(yi)
            if not x:
                continue

            n = ns[0]
            a = self.runs[n]
            a["label"] = ",".join([str(ni) for ni in ns])

            x, y = np.array(x), np.array(y)
            y = y[np.argsort(x)]
            x = np.sort(x)
            if self.E0 is None:
                en = a["dcm_ener"]  # EI incident energy
            else:
                en = self.E0
            x -= en

            a["E0"] = en
            a.pop("oneshot", None)

            header = "experiment: {0}\n".format(self.exp)
            header += "run: {0}\n".format(n)
            header += "date: {}\n".format(a["date"])
            header += "command: {0}\n".format(" ".join(a["command"]))
            header += "dcm_ener: {0}\n".format(a["dcm_ener"])
            header += "rixs_ener: {0}\n".format(a["rixs_ener"])
            if "t_coldhead" in a:
                header += "t_coldhead: {0}\n".format(a["t_coldhead"])
            if "t_sample" in a:
                header += "t_sample: {0}\n".format(a["t_sample"])
            header += "rixs_th: {0}\n".format(a["rixs_th"])
            header += "rixs_chi: {0}\n".format(a["rixs_chi"])
            header += "rixs_sam_x: {0}\n".format(a["rixs_sam_x"])
            header += "rixs_sam_y: {0}\n".format(a["rixs_sam_y"])
            header += "rixs_sam_z: {0}\n".format(a["rixs_sam_z"])
            header += "q_hkl: {0:.4f} {1:.4f} {2:.4f}\n".format(
                a["qh"], a["qk"], a["ql"]
            )
            header += "det_threshold: {0}\n".format(a["threshold"])
            header += "det_cutoff: {0}\n".format(a["cutoff"])
            header += "det_factor: {0}\n".format(a["detfac"])
            header += "det_roix: {0}\n".format(roix)
            header += "det_roiy: {0}\n".format(roiy)
            header += "E0_ypixel: {0}\n".format(y0)
            header += "E0_offset: {0}\n".format(en)

            if self.photon_counting:
                savefile = "{0}/{1}_pc_{2:05d}.txt".format(
                    self.savedir_dat, self.exp, n
                )
            else:
                savefile = "{0}/{1}_{2:05d}.txt".format(self.savedir_dat, self.exp, n)
            np.savetxt(
                savefile,
                np.array([x, y]).T,
                header=header + "\n{0:>24}{1:>24}".format(a["auto"], "counts"),
            )

            y = y / self.photon_factor
            if self.photon_counting and self.max_events:
                x = np.delete(x, np.where(y > self.max_events + 0.5))
                y = np.delete(y, np.where(y > self.max_events + 0.5))
                for i in range(self.max_events, 1, -1):
                    cnts = np.where(np.logical_and(y >= i - 0.5, y < i + 0.5))
                    y[cnts] /= i
                    if i > 2:
                        y = np.append(y, np.tile(y[cnts], i - 1))
                        x = np.append(x, np.tile(x[cnts], i - 1))
                    else:
                        y = np.append(y, y[cnts])
                        x = np.append(x, x[cnts])
                y = y[np.argsort(x)]
                x = np.sort(x)
            if bins:
                x, y, e = binning(x, y, bins, self.photon_counting)
            else:
                e = np.sqrt(y)
            y[~np.isfinite(y)] = 0
            a["x"], a["y"], a["e"] = x, y, e

            if fit:
                a["xf"], a["yf"], a["p"] = peak_fit(x, y)
                report = "#{0:<4} (bin: {1})  ".format(n, bins)
                report += "cen:{0:8.4f}   ".format(a["p"][2])
                report += "amp:{0:6.2f}   ".format(a["p"][0])
                report += "fwhm:{0:6.3f}   ".format(a["p"][1] * 2)
                report += "fra:{0:4.1f}   ".format(a["p"][3])
                report += "bg:{0:6.3f}".format(a["p"][3])
                print(report)
            else:
                a["p"] = False

            header += "bin_size: {0}\n".format(bins)
            header += "\n{0:>24}{1:>24}{2:>24}".format(a["auto"], "counts", "stderr")
            if bins < 5:
                savefile = "{0}/{1}_{2:05d}_b{3:.1f}meV.txt".format(
                    self.savedir_con, self.exp, n, bins * 1000
                )
            else:
                savefile = "{0}/{1}_{2:05d}_b{3}.txt".format(
                    self.savedir_con, self.exp, n, bins
                )
            np.savetxt(savefile, np.array([x, y, e]).T, header=header)

    def plot(
        self,
        run_nos,
        ax=None,
        step="run",
        labels=None,
        sort=False,
        rev=False,
        norm=False,
        ysca=None,
        ystp=0,
        yoff=None,
        xoff=None,
        show_fit=True,
        stderr=False,
        cmap=None,
        fmt="-",
        lw=1,
        vline=[0],
        leg=0,
        title=None,
        savefig=True,
        xlim=None,
        ylim=None,
    ):

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        run_nos = [n[0] if isinstance(n, list) else n for n in run_nos]
        runs = [
            self.runs[n] for n in run_nos if n in self.runs and self.runs[n] is not None
        ]
        if not runs:
            return

        if sort:
            runs.sort(key=lambda x: x[step])
        if rev:
            runs = runs[::-1]

        if labels:
            if not isinstance(labels, (list, tuple)):
                labels = [labels]

        if ax is None:
            _, ax = plt.subplots(figsize=(6, 5), constrained_layout=True)
        if cmap:
            if isinstance(cmap, str):
                cmap = plt.get_cmap(cmap)

        for i, a in enumerate(runs):

            xf, yf, p = False, False, False
            if "x" in a:
                x, y, e = deepcopy(a["x"]), deepcopy(a["y"]), deepcopy(a["e"])
                if a["p"] is not False:
                    xf, yf, p = a["xf"], a["yf"], a["p"]
            else:
                print("#{0}: nothing to plot".format(a["run"]))
                continue

            if xoff is not None:
                if isinstance(xoff, list):
                    x += xoff[i]
                else:
                    x += xoff

            if norm:
                if norm == "time":
                    maxy = a["time"]
                if isinstance(norm, (tuple, list)):
                    maxy = np.mean(y[(x > min(norm)) & (x < max(norm))])
                else:
                    maxy = max(y)
                y = y / maxy
                if e is not False:
                    e = e / maxy
                if p is not False:
                    yf = yf / maxy

            if ysca is not None:
                if isinstance(ysca, list):
                    ys = ysca[i]
                else:
                    ys = ysca
                y *= ys
                if e is not False:
                    e *= ys
                if p is not False:
                    yf *= ys

            if yoff is not None:
                if isinstance(yoff, list):
                    yo = yoff[i]
                else:
                    yo = yoff
                y += yo
                if p is not False:
                    yf += yo

            if cmap:
                c = cmap(i / len(runs))
            else:
                c = None

            label = "#{0}".format(a["label"])
            if labels:
                label += " {0}".format(labels[i])
            elif step == "T":
                label += " {:.0f}K".format(a[step])
            elif step == "rixs_th":
                label += " th: {:.2f}".format(a[step])
            elif step == "hkl":
                label += " {:.2f} {:.2f} {:.2f}".format(a["qh"], a["qk"], a["ql"])
            elif step != "run":
                try:
                    label += " {}: {:.3f}".format(step, a[step])
                except TypeError:
                    label += " {}: {}".format(step, a[step])

            (l,) = ax.plot(x, y + i * ystp, fmt, color=c, lw=lw, label=label)

            if stderr and e is not None and not norm:
                ax.errorbar(x, y + i * ystp, e, fmt="none", color=l.get_color(), lw=lw)

            if show_fit and p is not False:
                amp, sig, cen, fra, bgnd = p

                ax.plot(xf, yf + i * ystp, color=l.get_color(), lw=0.5)

                if len(runs) == 1 and len(ax.texts) == 0:
                    fr = "amp: {:.2f}\nfwhm: {:.3f}\ncen: {:.4f}".format(
                        amp, sig * 2, cen
                    )
                    ax.text(
                        0.05,
                        0.95,
                        fr,
                        va="top",
                        linespacing=1.3,
                        transform=ax.transAxes,
                    )

        ax.minorticks_on()

        if norm == "time":
            ax.set_ylabel("Intensity (per second)")
        elif norm:
            ax.set_ylabel("Intensity (normalised)")
        else:
            ax.set_ylabel("Intensity")

        if "E0" in a:
            ax.set_xlabel("Energy Transfer (eV)")
        elif a["oneshot"]:
            ax.set_xlabel("y-pixel")
        else:
            ax.set_xlabel(a["auto"])

        if leg is not False:
            if len(runs) > 20:
                ncol = 2
            else:
                ncol = 1
            ax.legend(
                loc=leg,
                handlelength=1.5,
                labelspacing=0.3,
                handletextpad=0.5,
                ncol=ncol,
                fontsize="small",
            )

        if "E0" not in a:
            vline = False
        if vline is not False:
            if not isinstance(vline, (tuple, list)):
                vline = [vline]
            for v in vline:
                ax.axvline(v, color="k", lw=0.5)

        if title:
            ax.set_title(title)
        if xlim:
            ax.set_xlim(*xlim)
        if ylim:
            ax.set_ylim(*ylim)

        if savefig:
            if not isinstance(savefig, str):
                sc = "_".join(str(n) for n in run_nos)
                savefig = "s{}_{}".format(sc, a["auto"])
            plt.savefig("{}/{}.pdf".format(self.savedir_fig, savefig), dpi=300)

    def check_run(
        self,
        run_no,
        hist=False,
        no=0,
        vmin=0,
        vmax=10,
        interp="hanning",
        photon_counting=False,
    ):

        """Step through each detector image of a run.
        run -- run number
        no -- starting step number (default is the first)
        hist -- plot histogram rather than detector map
        vmin -- colourmap minimum
        vmax -- colourmap maximum
        interp -- interpolation mode for image plot ('nearest' to disable)
        photon_counting -- check photon counting algorithm
        """

        self.load(run_no)
        a = self.runs[run_no]

        pnts = a["pnts"]
        to = a["threshold"] - a["detfac"]
        co = a["cutoff"] - a["detfac"]

        i = {}
        i["idx"] = no
        fig, i["ax"] = plt.subplots()

        imdat = a["img"][no]
        imgdim = imdat.shape

        if hist:
            b, c = np.histogram(imdat, bins=range(to, co, 1))
            i["ax"].bar(c[:-1], b)
        elif photon_counting:
            imdat, _ = scipy.ndimage.label(imdat)
            # cmap = plt.get_cmap('prism')
            vals = np.linspace(0, 1, 256)
            np.random.shuffle(vals)
            cmap = plt.cm.colors.ListedColormap(plt.get_cmap("jet")(vals))
            cmap.set_under("k")
            i["im"] = i["ax"].imshow(
                imdat, origin="lower", cmap=cmap, interpolation="nearest", vmin=0.1
            )
        else:
            cmap = plt.get_cmap("bone_r")
            i["im"] = i["ax"].imshow(
                imdat,
                origin="lower",
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                interpolation=interp,
            )
        i["ax"].set_title("#{} no {}".format(run_no, i["idx"]))

        def do_plot(i):
            try:
                imdat = a["img"][i["idx"]]
            except IndexError:
                imdat = np.zeros(imgdim)
                print("check run: no tiff for step {}".format(i["idx"]))
            if hist:
                b, c = np.histogram(imdat, bins=range(to, co, 1))
                i["ax"].cla()
                i["ax"].bar(c[:-1], b)
            else:
                if photon_counting:
                    imdat, _ = scipy.ndimage.label(imdat)
                i["im"].set_data(imdat)
            i["ax"].set_title("#{} no {}".format(run_no, i["idx"]))
            plt.draw()

        def press(event):
            if event.key == "left":
                if i["idx"] > 0:
                    i["idx"] -= 1
                    do_plot(i)
            elif event.key == "right":
                if i["idx"] < pnts - 1:
                    i["idx"] += 1
                    do_plot(i)

        fig.canvas.mpl_connect("key_press_event", press)

    def calc_distortion(
        self, run_no, slices=8, oneshot=True, no=0, plot=False, vmin=0, vmax=10
    ):

        self.load(run_no)
        a = self.runs[run_no]
        if "img" not in a:
            print("detector images not loaded")
            return

        if oneshot:
            img = np.atleast_3d(np.array(a["img"]))
            img = np.sum(img, axis=0)
        else:
            img = a["img"][no]
        img = img[self.roiy[0] : self.roiy[1], self.roix[0] : self.roix[1]]

        y = np.sum(img, axis=1)
        x = np.arange(self.roiy[0], self.roiy[1])
        _, _, pinit = peak_fit(x, y)
        y0 = int(round(pinit[2]))
        print("fitted y0: {}".format(y0))
        print("initial fwhm: {:.4f}".format(pinit[1] * 2))

        slice_width = img.shape[1] / slices
        shift = []
        regions = []

        for i in range(slices):
            c1, c2 = int(i * slice_width), int(i * slice_width + slice_width)
            yi = np.sum(img[:, c1:c2], axis=1)
            try:
                _, _, pi = peak_fit(x, yi)
                cen = int(round(pi[2]))
            except RuntimeError:
                cen = y0
            shift.append(y0 - cen)
            regions.append([c1, c2])

        self.corr_shift = shift
        self.corr_regions = regions

        imgcorr = deepcopy(img)
        for sh, (c1, c2) in zip(shift, regions):
            yi = np.sum(imgcorr[:, c1:c2], axis=1)
            imgcorr[:, c1:c2] = np.roll(imgcorr[:, c1:c2], sh, axis=0)

        ycorr = np.sum(imgcorr, axis=1)
        _, _, pfinal = peak_fit(x, ycorr)
        print("final fwhm: {:.4f}".format(pfinal[1] * 2))

        if plot:
            _, ax = plt.subplots(1, 3, figsize=(10, 4), constrained_layout=True)
            ax[0].plot(x, y, lw=0.5)
            ax[0].plot(x, ycorr, lw=0.5)

            ax[1].imshow(
                img,
                origin="lower",
                cmap=plt.get_cmap("bone_r"),
                aspect="auto",
                extent=(self.roix[0], self.roix[1], self.roiy[0], self.roiy[1]),
                vmin=vmin,
                vmax=vmax,
                interpolation="hanning",
            )

            ax[2].imshow(
                imgcorr,
                origin="lower",
                cmap=plt.get_cmap("bone_r"),
                aspect="auto",
                extent=(self.roix[0], self.roix[1], self.roiy[0], self.roiy[1]),
                vmin=vmin,
                vmax=vmax,
                interpolation="hanning",
            )

            ax[1].axhline(y0, color="#F012BE", lw=0.5)
            ax[2].axhline(y0, color="#F012BE", lw=0.5)
            for sh, (c1, c2) in zip(shift, regions):
                ax[1].axvline(c1 + self.roix[0], color="#F012BE", lw=0.5)
                ax[2].axvline(c1 + self.roix[0], color="#F012BE", lw=0.5)
